import numpy as np

from bioevents import event_handling, hypnogram


def simulate_hypnogram(series_duration, num_cycles=5, avg_duration=5, seed=0):
    rng = np.random.default_rng(seed)
    classes = hypnogram.SleepStages
    stack = {k: [] for k in classes}
    c = 5
    stack_duration = 0
    cycles = (
        np.abs(np.sin(np.pi * num_cycles * np.linspace(0, 1, series_duration))) * np.ptp(classes)
        - 0.2
    )
    while stack_duration < series_duration:
        dur = int(
            np.clip(
                rng.normal(avg_duration, avg_duration / 3),
                1,
                series_duration - stack_duration,
            )
        )
        stack[c].append(event_handling.Event(stack_duration, stack_duration + dur))
        c = max(classes) - cycles[stack_duration] + rng.normal() / 3
        c = int(np.clip(c, min(classes), max(classes)))
        stack_duration += dur
    stack = hypnogram.Hypnogram(
        {
            k: event_handling.EventSeries(events, duration=series_duration)
            for k, events in stack.items()
            if events
        }
    )
    return stack



import abc
import datetime
from enum import IntEnum
from typing import Optional

import matplotlib.pyplot as plt
import numpy as np
from pydantic import (
    BaseModel,
    NonNegativeFloat,
    NonNegativeInt,
    confloat,
    root_validator,
)

from bioevents import event_handling

WAKE_PERSISTENCE_THRESHOLD_IN_MINUTES = 2


class SleepStages(IntEnum):
    W = 5
    REM = 4
    N1 = 3
    N2 = 2
    N3 = 1


class Labels(IntEnum):
    W = SleepStages.W.value
    REM = SleepStages.REM.value
    N1 = SleepStages.N1.value
    N2 = SleepStages.N2.value
    N3 = SleepStages.N3.value
    MISSING = event_handling.MissingDataLabels.MISSING.value
    UNSCORABLE = event_handling.MissingDataLabels.UNSCORABLE.value
    ARTIFACT = event_handling.MissingDataLabels.ARTIFACT.value
    DISAGREE = event_handling.MissingDataLabels.DISAGREE.value


class HypnogramReport(BaseModel, abc.ABC):
    """Used for the EDS reports"""

    duration_minutes: NonNegativeFloat
    latency_minutes_rem: Optional[NonNegativeFloat] = None
    latency_minutes_n1: Optional[NonNegativeFloat] = None
    latency_minutes_n2: Optional[NonNegativeFloat] = None
    latency_minutes_n3: Optional[NonNegativeFloat] = None
    latency_minutes_persistent_sleep: Optional[NonNegativeFloat] = None
    latency_minutes_rem_with_persistence: Optional[NonNegativeFloat] = None
    latency_minutes_sleep_onset: Optional[NonNegativeFloat] = None
    latency_minutes_sleep_onset_unequivocal: Optional[NonNegativeFloat] = None
    total_minutes_rem: NonNegativeFloat
    total_minutes_n1: NonNegativeFloat
    total_minutes_n2: NonNegativeFloat
    total_minutes_n3: NonNegativeFloat
    total_minutes_nrem: NonNegativeFloat
    total_minutes_sleep: NonNegativeFloat
    efficiency_rem: confloat(ge=0, le=1)
    efficiency_rem_first_two_hours: confloat(ge=0, le=1)
    efficiency_sleep: confloat(ge=0, le=1)
    event_count_rem: NonNegativeInt
    event_count_wake: NonNegativeInt
    event_count_wake_persistent: NonNegativeInt
    wake_after_sleep_onset_minutes: Optional[NonNegativeFloat] = None

    def __str__(self):
        rows = [
            f"Latency to REM:                       {self.latency_minutes_rem:.1f} minutes",
            f"Latency to N1:                        {self.latency_minutes_n1:.1f} minutes",
            f"Latency to N2:                        {self.latency_minutes_n2:.1f} minutes",
            f"Latency to N3:                        {self.latency_minutes_n3:.1f} minutes",
            f"Latency to persistent sleep:          {self.latency_minutes_persistent_sleep:.1f} minutes",
            f"Latency to REM from persistent sleep: {self.latency_minutes_rem_with_persistence:.1f} minutes",
            f"Latency to sleep onset (SOL)          {self.latency_minutes_sleep_onset:.1f} minutes",
            f"Latency to sleep onset (SOL-U)        {self.latency_minutes_sleep_onset_unequivocal:.1f} minutes",
            f"Total time in REM:                    {self.total_minutes_rem:.1f} minutes",
            f"Total time in N1:                     {self.total_minutes_n1:.1f} minutes",
            f"Total time in N2:                     {self.total_minutes_n2:.1f} minutes",
            f"Total time in N3:                     {self.total_minutes_n3:.1f} minutes",
            f"Total time in non-REM sleep:          {self.total_minutes_nrem:.1f} minutes",
            f"Total time asleep:                    {self.total_minutes_sleep:.1f} minutes",
            f"Efficiency of REM:                    {100 * self.efficiency_rem:.1f} %",
            f"Efficiency of REM (first two hours):  {100 * self.efficiency_rem_first_two_hours:.1f} %",
            f"Efficiency of sleep:                  {100 * self.efficiency_sleep:.1f} %",
            f"Number of REM cycles:                 {self.event_count_rem}",
            f"Number of awakenings:                 {self.event_count_wake}",
            f"Number of awakenings >120s:           {self.event_count_wake_persistent}",
        ]
        return "\n".join(rows)

    @root_validator(allow_reuse=True)
    def check_total_times_add_up(cls, values):
        rem = values.get("total_minutes_rem")
        n1 = values.get("total_minutes_n1")
        n2 = values.get("total_minutes_n2")
        n3 = values.get("total_minutes_n3")
        nrem = values.get("total_minutes_nrem")
        sleep = values.get("total_minutes_sleep")
        if not np.isclose(n1 + n2 + n3, nrem):
            raise ValueError(
                "Total minutes in non-REM sleep doesn't match individual stage times."
            )
        if not np.isclose(nrem + rem, sleep):
            raise ValueError("Total minutes in sleep doesn't match REM + NREM times.")
        return values

    @root_validator(allow_reuse=True)
    def check_efficiencies(cls, values):
        rem = values.get("efficiency_rem")
        rem_time = values.get("total_minutes_rem")
        sleep = values.get("efficiency_sleep")
        sleep_time = values.get("total_minutes_sleep")
        dur = values.get("duration_minutes")
        if not np.isclose(rem_time / dur, rem):
            raise ValueError("REM efficiency and total time don't match.")
        if not np.isclose(sleep_time / dur, sleep):
            raise ValueError("Sleep efficiency and total time don't match.")
        return values

    @root_validator(allow_reuse=True)
    def check_awakenings(cls, values):
        total = values.get("event_count_wake")
        persistent = values.get("event_count_wake_persistent")
        if total < persistent:
            raise ValueError("Persistent event count is greater than total count!")
        return values

    @root_validator(allow_reuse=True)
    def check_no_rem(cls, values):
        value_if_no_rem = {
            "efficiency_rem": 0,
            "total_minutes_rem": 0,
            "event_count_rem": 0,
            "latency_minutes_rem": None,
        }
        no_rem_indicators = [values[k] == value_if_no_rem[k] for k in value_if_no_rem]
        if not any(no_rem_indicators):
            return values
        if all(no_rem_indicators):
            return values
        raise ValueError("Some, but not all, REM metrics indicate REM is present!")


class Hypnogram(event_handling.EventClassSeries):
    def __init__(self, event_stack):
        super().__init__(event_stack)
        # make sure that we fill in any missing sleep stages with empty event series
        self.update(
            {
                c: event_handling.EventSeries(
                    [], duration=self.duration, sampling_rate=self.sampling_rate
                )
                for c in list(SleepStages)
                if c not in self.classes
            }
        )

    @classmethod
    def read_json(cls, filepath, intenum=Labels):
        return super().read_json(filepath, intenum)

    def total_sleep_time(self, bedtime_series: event_handling.EventSeries = None) -> float:
        """Find total sleep time (in samples) from a hypnogram

        Parameters
        ----------
        bedtime_series : Optional[event_handling.EventSeries]
            An event series derived from clinical lights off/on annotations.
            In the clinical setting, this is the period of time when subjects are expected to attempt sleep.
            If given, only compute sleep time that occurs during bedtime.

        Notes
        -----
        This metric is traditionally reported in minutes. Please take note of the sample_rate to convert if needed.
        """
        if bedtime_series is not None:
            return sum([self.trim(ev.on, ev.off).total_sleep_time() for ev in bedtime_series])
        return self.duration - self.time_in_stage(SleepStages.W)

    def total_recording_time(self, bedtime_series: event_handling.EventSeries = None) -> float:
        """Find total recording time, i.e., total time in bed, from a hypnogram (in samples)

        Parameters
        ----------
        bedtime_series : Optional[event_handling.EventSeries]
            An event series derived from clinical lights off/on annotations.
            In the clinical setting, this is the period of time when subjects are expected to attempt sleep.
            If given, only compute recording time that occurs during bedtime.

        Notes
        -----
        This metric is traditionally reported in minutes. Please take note of the sample_rate to convert if needed.
        """
        if bedtime_series is None:
            return self.duration
        return sum([ev.duration for ev in bedtime_series])

    def sleep_efficiency_percent(self, bedtime_series: event_handling.EventSeries = None) -> float:
        eff = self.total_sleep_time(bedtime_series) / self.total_recording_time(bedtime_series)
        return 100 * eff

    # NOTE: the hypnogram needs to be trimmed
    def sleep_onset_latency(self, n1_tolerance: float = 0) -> Optional[float]:
        """Find sleep onset latency from a hypnogram (in samples). Assume the start of the hypnogram is the start of
        bedtime (light off).

        Parameters
        ----------
        n1_tolerance : float, default 0
            If 0 (default), return latency to sleep onset as the first non-Wake stage (used for overnight sleep)
            If positive, units are samples, and more rules are followed (see Notes)

        Returns
        -------
        sleep_onset_latency : Optional(float)
            Time (in samples) from beginning of this hypnogram until the sleep onset.
            See Notes for more information on the definition of sleep onset.

        Notes
        -----
        This metric is traditionally reported in minutes. Please take note of the sample_rate to convert if needed.
        There are multiple ways to report SOL:

        1. The start of first scored epoch of any stage of sleep.

            Typically used in overnight PSG.

        2. The start of the first epoch of "unequivocal sleep".

            Typically used in MWT for assessment of daytime
            sleepiness. "Unequivocal sleep" is "defined as 3 consecutive epochs of stage 1 sleep,
            or 1 epoch of any other stage of sleep" (see Reference #1). Epochs, in this case, are 30 seconds long.
            Thus, the first epoch of any of the following would be considered unequivocal sleep onset:

                - [N1, N1, N1, ...] three epochs of N1
                - [N1, N1, S, ...] two epochs of N1, followed immediately by any other stage of sleep
                - [N1, S, ...] one epoch of N1, followed immediately by any other stage of sleep
                - [S, ...] one epoch of any other stage of sleep

            For algorithmic purposes, this reduces to the observation that unequivocal sleep occurs at the onset of ANY
            sleep other than short, N1-only sleep episodes lasting less than 90 seconds.

        References
        ----------
        1. Clinical study using SOL as a primary endpoint: https://clinicaltrials.gov/ct2/show/NCT03522506
        """
        asleep = ~self[SleepStages.W]

        if not len(asleep):
            return None

        if n1_tolerance == 0:
            # return onset of first non-wake event
            return asleep[0].on

        # first remove n1 events shorter than the tolerance from the "asleep" time series
        n1 = self[SleepStages.N1]
        n1_exclude = event_handling.EventSeries(
            [ev for ev in n1 if ev.duration <= n1_tolerance],
            duration=self.duration,
            sampling_rate=self.sampling_rate,
        )
        asleep -= n1_exclude

        if not len(asleep):
            return None

        # IF there's ANY N1 event that preceded the first sleep event, we'll use it as the onset
        first_sleep_event = asleep[0]
        for n1_event in n1:
            if n1_event.off == first_sleep_event.on:
                first_sleep_event = n1_event
        return first_sleep_event.on

    def wake_after_sleep_onset(self) -> Optional[float]:
        """Compute WASO (wake after sleep onset) from a hypnogram. Useful for overnight sleep

        Returns
        -------
        WASO : Optional(float)
            If SOL is None, returns None.
            If SOL is not None, returns wake after sleep onset (in samples).

        Notes
        -----
        This metric is traditionally reported in minutes. Please take note of the sample_rate to convert if needed.
        """
        SOL = self.sleep_onset_latency(n1_tolerance=0)
        if SOL is None:
            return None
        TRT = self.total_recording_time()
        TST = self.total_sleep_time()
        WASO = TRT - TST - SOL
        if WASO < -0.1:
            raise RuntimeError("WASO is negative and exceeds precision error. This is a bug.")
        WASO = max(WASO, 0.0)  # avoids precision errors
        return WASO

    def time_in_stage(self, stage: SleepStages) -> float:
        """Total time spent in the given stage, in samples."""
        return sum([ev.duration for ev in self[stage]])

    def latency_to_first_persistent_sleep(self) -> Optional[float]:
        """Latency to first epoch of persistent (10 minutes) sleep.

        Returns
        -------
        latency : Optional(float)
            If no persistent sleep occurs, returns None.
            If persistent sleep occurs, returns latency (in samples).

        Notes
        -----
        This metric is traditionally reported in minutes. Please take note of the sample_rate to convert if needed.
        """
        asleep = ~self[SleepStages.W]

        persistence_threshold_seconds = datetime.timedelta(minutes=10).seconds
        persistence_threshold_samples = persistence_threshold_seconds * asleep.sampling_rate

        persistent_sleep = [ev for ev in asleep if ev.duration >= persistence_threshold_samples]
        if len(persistent_sleep):
            return persistent_sleep[0].on
        return None

    def latency_to_stage(self, stage: SleepStages) -> Optional[float]:
        """Time from lights off to first epoch of the given stage

        Parameters
        ----------
        stage : SleepStages

        Returns
        -------
        latency : Optional[float]
            If stage does not occur, returns None. Otherwise, returns latency (in samples).

        Notes
        -----
        This metric is traditionally reported in minutes. Please take note of the sample_rate to convert if needed.
        """
        events = self[stage]
        if len(events) != 0:
            return events[0].on
        return None

    def generate_report(self) -> HypnogramReport:
        self_in_minutes = self.resample(1 / 60)
        self_first_two_hours = (
            self_in_minutes if self_in_minutes.duration < 120 else self_in_minutes.trim(end=120)
        )
        stages_nrem = [SleepStages.N1, SleepStages.N2, SleepStages.N3]

        sol = self_in_minutes.sleep_onset_latency(n1_tolerance=0.0)
        waso = self_in_minutes.wake_after_sleep_onset()

        # Although it might be safe to assume we have resampled 30-seconds epochs,
        # let's tolerate n1 episodes up to 2.5 30-s epochs.
        sol_unequivocal = self_in_minutes.sleep_onset_latency(n1_tolerance=1.25)

        latency_persistent_sleep = self_in_minutes.latency_to_first_persistent_sleep()
        if latency_persistent_sleep is not None:
            self_from_persistent = self_in_minutes.trim(
                latency_persistent_sleep, self_in_minutes.duration
            )
            latency_rem_from_persistent = self_from_persistent.latency_to_stage(SleepStages.REM)
        else:
            latency_rem_from_persistent = None
        wake_time = self_in_minutes.time_in_stage(SleepStages.W)
        rem_time = self_in_minutes.time_in_stage(SleepStages.REM)
        rem_time_first_two_hours = self_first_two_hours.time_in_stage(SleepStages.REM)
        persistent_wake_events = [
            ev
            for ev in self_in_minutes[SleepStages.W]
            if ev.duration > WAKE_PERSISTENCE_THRESHOLD_IN_MINUTES
        ]
        report = HypnogramReport(
            duration_minutes=self_in_minutes.duration,
            latency_minutes_rem=self_in_minutes.latency_to_stage(SleepStages.REM),
            latency_minutes_n1=self_in_minutes.latency_to_stage(SleepStages.N1),
            latency_minutes_n2=self_in_minutes.latency_to_stage(SleepStages.N2),
            latency_minutes_n3=self_in_minutes.latency_to_stage(SleepStages.N3),
            latency_minutes_persistent_sleep=latency_persistent_sleep,
            latency_minutes_rem_with_persistence=latency_rem_from_persistent,
            latency_minutes_sleep_onset=sol,
            latency_minutes_sleep_onset_unequivocal=sol_unequivocal,
            total_minutes_rem=rem_time,
            total_minutes_n1=self_in_minutes.time_in_stage(SleepStages.N1),
            total_minutes_n2=self_in_minutes.time_in_stage(SleepStages.N2),
            total_minutes_n3=self_in_minutes.time_in_stage(SleepStages.N3),
            total_minutes_nrem=sum([self_in_minutes.time_in_stage(s) for s in stages_nrem]),
            total_minutes_sleep=self_in_minutes.duration - wake_time,
            efficiency_rem=rem_time / self_in_minutes.duration,
            efficiency_rem_first_two_hours=rem_time_first_two_hours / 120,
            efficiency_sleep=1 - wake_time / self_in_minutes.duration,
            event_count_rem=len(self_in_minutes[SleepStages.REM]),
            event_count_wake=len(self_in_minutes[SleepStages.W]),
            event_count_wake_persistent=len(persistent_wake_events),
            wake_after_sleep_onset_minutes=0.0 if waso is None else waso,
        )
        return report

    def plot(self, unit="seconds", ax=None, color=None, alpha=1, label=None, **plt_kwargs):
        """Stacks individual series onto the same plot

        Parameters
        ----------
        unit : str, Optional[default="seconds"]
            Must be one of the keyword args to datetime.timedelta
            This will affect the x-axis label's units.
        ax : plt.Axes, Optional
        color : plt.colors[Optional]
            Color to use for the hypnogram and any non-sleep labels
        alpha : float, Optional[default=1]
            Alpha to use for hypnogram. A gradient thereof will be used for any non-sleep labels.
        label : str, Optional
            Label to use for this hypnogram. Useful when plotting multiple hypnograms on the given axis "ax"
        plt_kwargs : kwargs
            Any additional keyword args meant for plt.plot AND plt.fill_betweenx

        Notes
        -----
        This function is better aligned with the format of a hypnogram plot that is expected in sleep science.
        """
        if ax is None:
            fig, ax = plt.subplots(figsize=(10, 0.5 * len(self)))

        # handle unit conversion
        try:
            unit_hz = 1 / datetime.timedelta(**{unit: 1}).total_seconds()
        except TypeError:
            raise ValueError("Unit argument must be one of those accepted by datetime.timedelta!")
        resampled = self.resample(unit_hz)
        ax.set_xlabel(f"Time ({unit})")

        events_and_labels = []
        for key, values in resampled.items():
            event_label = key if key in SleepStages else np.nan
            for event in values:
                events_and_labels.append((event, event_label))
        events_and_labels = sorted(events_and_labels, key=lambda ev: ev[0].on)
        x, event_labels = [], []
        for event, event_label in events_and_labels:
            x.append(event.on)
            event_labels.append(event_label)
            x.append(event.off)
            event_labels.append(event_label)

        # "typical" hypnogram
        p = ax.plot(x, event_labels, color=color, alpha=alpha, label=label, **plt_kwargs)
        stages = [s for s in resampled if isinstance(s, SleepStages)]
        ax.set_yticks([s.value for s in stages])
        ax.set_yticklabels([s.name for s in stages])

        # pull out the color used for the hypnogram itself in case we need to reuse it
        color = p[0].get_color() if color is None else color

        missing_data_labels = {
            event_label: events
            for event_label, events in resampled.items()
            if isinstance(event_label, event_handling.MissingDataLabels)
        }

        if len(missing_data_labels):
            top = max(event_labels)
            for k, v in missing_data_labels.items():
                for on, off in v:
                    ax.fill_betweenx(
                        y=[0, top],
                        x1=on,
                        x2=off,
                        color=color,  # reuse color from line
                        alpha=alpha / (1 - k),  # make an alpha gradient by enumeration
                        label=k if label is None else f"{label}: ({k.name})",
                        **plt_kwargs,  # any additional kwargs we'd like to reuse
                    )

        # reset the legend to remove any duplicate labels
        handles, event_labels = ax.get_legend_handles_labels()
        by_label = dict(zip(event_labels, handles))
        ax.legend(by_label.values(), by_label.keys(), frameon=False, bbox_to_anchor=(1, 1))

        ax.set_ylim(ax.get_ylim())
        ax.set_xlabel(f"Time ({unit})")



from __future__ import annotations

import json
from decimal import Decimal
from enum import IntEnum
from typing import Dict, Optional, ValuesView

import matplotlib.pyplot as plt
from matplotlib.colors import LinearSegmentedColormap
import numpy as np
import pandas as pd
from bioevents.colors import Colors

DEFAULT_SAMPLING_RATE_HZ = 1.0


class OverlapTolerances:
    def __init__(self, diff_on=np.inf, diff_off=np.inf, ratio_on=np.inf, ratio_off=np.inf):
        """Useful for storing various tolerances employed by the Event.overlaps function

        Parameters
        ----------
        diff_on, diff_off, ratio_on, ratio_off : int or float
            Must be non-negative floats.
            diff_on, diff_off: maximum acceptable difference in on or off timestamps
            ratio_on, ratio_off: maximum acceptable ratio between
                difference in on or off timestamps and reference duration

        Notes
        -----
        This may be a somewhat odd architecture for a collection of configurations.
        However, note that it enables several unique handling operations via customized built-in functions.
        In addition, one may pre-configure this object and propagate it through an EventStack or EventSeries.

        All timestamp diffs are unsigned, meaning tolerances apply symmetrically about the reference timestamp.

        Note that the default configuration allows for events with ANY overlapping samples to overlap.
        This is consistent with "Seizure Detection" (Scheuer et al., 2021)

        References
        ----------
        "Spike Detection", Mark L. Scheuer et al., 2016
            "Spikes marked by different readers that occurred within 200 milliseconds of one another
            were considered to be the same event.""
        "Seizure Detection", Mark L. Scheuer et al., 2021
            "Seizure marks from a reader pair were designated a match (true positive) if they overlapped"

        See Also
        --------
        Event.overlaps
        EventSeries.set_tolerances
        EventStack.set_tolerances
        """
        self.diff_on = diff_on
        self.diff_off = diff_off
        self.ratio_on = ratio_on
        self.ratio_off = ratio_off

    def __repr__(self):
        """Allows for quick inspection"""
        return str(dict(self))

    def __iter__(self):
        """Allows for dictionary composition"""
        for k in dir(self):
            if k.startswith("_"):
                continue
            yield k, getattr(self, k)

    def __setattr__(self, k, v):
        """Ensures all tolerances are non-negative numbers"""
        assert v >= 0, ValueError("Tolerance must be a non-negative number")
        super().__setattr__(k, v)


class Event:
    default_duration = 1

    def __init__(self, on, off=None, tolerances=None):
        """
        Container for timestamped, binary events.
        Also includes functionality for determining overlap with other events.
        """
        assert on >= 0, ValueError("Onset must be non-negative")
        off = on + self.default_duration if off is None else off
        assert off > on, ValueError("Non-positive duration calculated.")
        tolerances = OverlapTolerances() if tolerances is None else tolerances
        assert isinstance(tolerances, OverlapTolerances)
        self._on = on
        self._off = off
        self._tolerances = tolerances

    @classmethod
    def from_duration(cls, on, duration, tolerances=None):
        return cls(on, on + duration, tolerances)

    def __repr__(self):
        return str(dict(self))

    def __copy__(self):
        return Event(on=self.on, off=self.off, tolerances=self.tolerances)

    def __iter__(self):
        """Utility: TBD"""
        for k, v in zip(["on", "off"], [self.on, self.off]):
            yield k, v

    def __eq__(self, item):
        """Utility: TBD"""
        return self.on == item.on and self.off == item.off

    @property
    def on(self):
        return self._on

    @property
    def off(self):
        return self._off

    @property
    def duration(self):
        return self.off - self.on

    @property
    def tolerances(self):
        return self._tolerances

    @tolerances.setter
    def tolerances(self, vals):
        assert isinstance(vals, OverlapTolerances)
        self._tolerances = vals

    def overlaps(self, other):
        """
        Determines whether this event "overlaps" another, based on various criteria.

        Notes
        -----
        It's important to note that this overlap is not necessarily a "symmetrical" operation.
        Because the duration and tolerances of THIS event are used, self.overlaps(other) may differ
        from other.overlaps(self). This is by design, given that a "reference" or "truth" event should
        govern the acceptance criteria of "predicted" event annotations, and not the other way around.

        See Also
        --------
        OverlapTolerances
        """
        assert isinstance(other, Event)

        if self == other:
            return True

        # note -- this is rather arbitrary and could be reversed (see Notes)
        ref, test = self, other
        dur, tol = ref.duration, ref.tolerances

        transitions = [ref.on, test.on, ref.off, test.off]
        if max(transitions) - min(transitions) > (ref.duration + test.duration):
            # note that directly adjacent events will pass this condition
            return False  # there is no overlap at all

        diff = np.abs(ref.on - test.on)
        if diff > tol.diff_on:
            return False
        if (diff / dur) > tol.ratio_on:
            return False

        diff = np.abs(ref.off - test.off)
        if diff > tol.diff_off:
            return False
        if (diff / dur) > tol.ratio_off:
            return False

        return True

    def abuts(self, other):
        """Determines whether the onset of one event is the offset of the other"""
        return self.on == other.off or self.off == other.on


class EventSeries(list):
    def __init__(
        self,
        events,
        duration=None,
        sampling_rate: float = DEFAULT_SAMPLING_RATE_HZ,
    ):
        """Time series container for events.
        Provides various functionality for comparison between self and others.

        Parameters
        ----------
        events : list of Event
        duration : Optional(float or int)
            Duration of the entire time series.
            If omitted, the duration is assumed to be the offset of the latest event.
        sampling_rate : float, default=DEFAULT_SAMPLING_RATE_HZ
            Sampling rate in Hertz

        Notes
        -----
        Currently assumes the time series starts at timestamp 0.
        TODO: support non-zero start.
        TODO: support configurable sample rate
        """
        assert all(isinstance(item, Event) for item in events), ValueError("Non-event given.")
        super().__init__(events)
        if duration is not None:
            assert isinstance(duration, (float, int)) and (duration > 0)
        self._duration = duration
        self._sampling_rate = float(sampling_rate)
        self._resolve_events()

    def __contains__(self, item):
        assert isinstance(item, Event)
        return any(event.overlaps(item) for event in self)

    def __eq__(self, other):
        events_equal = super(EventSeries, self).__eq__(other)
        metadata_equal = self.is_compatible(other)
        return events_equal and metadata_equal

    def resample(self, sampling_rate):
        def convert(x):
            """Avoids numerical precision issues -- try running 3 * (1 / 10)"""
            return float(Decimal(x) / Decimal(self.sampling_rate) * Decimal(sampling_rate))

        return EventSeries(
            events=[Event(convert(ev.on), convert(ev.off), ev.tolerances) for ev in self],
            duration=convert(self.duration),
            sampling_rate=sampling_rate,
        )

    def trim(self, start: Optional[int] = None, end: Optional[int] = None):
        if start is None:
            start = 0
        elif start < 0:
            raise ValueError("Cannot trim outside the bounds of this EventSeries")

        if end is None:
            end = self.duration
        elif end > self.duration:
            raise ValueError("Cannot trim outside the bounds of this EventSeries")
        elif start >= end:
            raise ValueError("Trim start must be before trim end")

        trim_window = Event(start, end)
        events = []
        for event in self:
            if trim_window.abuts(event):
                continue
            if not trim_window.overlaps(event):
                continue
            events.append(
                Event(
                    max(0, event.on - start),
                    min(end, event.off) - start,
                    tolerances=event.tolerances,
                )
            )
        return EventSeries(events, duration=trim_window.duration, sampling_rate=self.sampling_rate)

    def is_compatible(self, item):
        """See if self vs. other functions can be used with this item"""
        assert isinstance(item, EventSeries), ValueError("EventSeries not given.")
        return self.sampling_rate == item.sampling_rate and self.duration == item.duration

    def spans_event(self, item):
        """See if the given event falls within the time bounds of this series"""
        assert isinstance(item, Event), ValueError("Event not given.")
        if self._duration is not None:
            return self._duration >= item.off

    def _assert_is_compatible(self, item):
        assert self.is_compatible(item), ValueError("EventSeries is not compatible.")

    def _assert_spans_event(self, item):
        assert self.spans_event(item), ValueError("Event exceeds duration")

    def __add__(self, item):
        """Defines built-in 'add' as a boolean intersection of compatible series"""
        self._assert_is_compatible(item)
        return EventSeries.from_bools(
            self.as_bools() + item.as_bools(), sampling_rate=self.sampling_rate
        )

    def __sub__(self, item):
        """Defines built-in 'add' as a boolean difference of compatible series"""
        self._assert_is_compatible(item)
        return EventSeries.from_bools(
            self.as_bools() & ~item.as_bools(), sampling_rate=self.sampling_rate
        )

    def __or__(self, item):
        """Defines built-in 'or' operator as a "smart" Event-wise intersection of compatible series"

        Notes
        -----
        -Overlapping events will be combined via logical "or"
        -Resulting events are "resolved"
        """
        self._assert_is_compatible(item)
        return self + item

    def __and__(self, item):
        """
        Notes
        -----
        -Overlap will be determined symmetrically by agreement between both items
        -Overlapping events will be combined via logical "or"
        -Resulting events are "resolved"
        """
        self._assert_is_compatible(item)
        events = []
        for a in self:
            for b in item:
                if a.overlaps(b) and b.overlaps(a):
                    events.append(Event(min(a.on, b.on), max(a.off, b.off)))
        return EventSeries(events, duration=self.duration, sampling_rate=self.sampling_rate)

    def __invert__(self):
        if len(self):
            events = []
            if self[0].on > 0:
                events.append(Event(on=0, off=self[0].on))
            for i in range(len(self) - 1):
                events.append(Event(on=self[i].off, off=self[i + 1].on))
            if self[-1].off < self.duration:
                events.append(Event(on=self[-1].off, off=self.duration))
        else:
            events = [Event(on=0, off=self.duration)]
        return EventSeries(events, duration=self.duration, sampling_rate=self.sampling_rate)

    def append(self, item):
        """Overrides built-in default to make sure the Event is concurrent with this series"""
        self._assert_spans_event(item)
        super().append(item)
        self._resolve_events()

    def extend(self, item):
        """Overrides built-in default to make sure the Events are concurrent with this series"""
        for e in item:
            self._assert_spans_event(e)
        super().extend(item)
        self._resolve_events()

    def _resolve_events(self):
        """Sort and combine overlapping events

        Notes
        -----
        Could use as/from bools to accomplish, but this can be much more efficient
        """
        if not len(self):
            return
        self.sort(key=lambda x: x.on, reverse=False)
        after = self.pop()
        final = []
        while len(self):
            before = self.pop()
            if before.off >= after.on:
                after = Event(before.on, max(before.off, after.off))
            else:
                final.append(after)
                after = before
        final.append(after)
        self.clear()
        super().extend(list(reversed(final)))

    def debounce(self, persistence_period: float, min_duration: float) -> EventSeries:
        """Debounce the given signal to remove short events and merge close ones

        Parameters
        ----------
        persistence_period : float
            merge any two events separated by less than this number (of samples)
        min_duration : float
            after merging close events, remove all events under this duration (in samples)

        Returns
        -------
        debounced : EventSeries

        References
        ----------
        Description of debouncing applied in engineering context: https://my.eng.utah.edu/~cs5780/debouncing.pdf
        """
        events_merged = []
        if len(self):
            # merge any events that are close to each other using a persistence period
            event = self[0]
            on, off = event.on, event.off
            for event in self[1:]:
                if event.on - off < persistence_period:
                    off = event.off  # merge events
                else:
                    # stash our new timestamps as a separate event
                    events_merged.append(Event(on, off))
                    on, off = event.on, event.off
            events_merged.append(Event(on, off))

        # keep only events that are longer than the minimum duration
        events_long_enough = [ev for ev in events_merged if ev.duration >= min_duration]

        return EventSeries(
            events_long_enough, duration=self.duration, sampling_rate=self.sampling_rate
        )

    @property
    def duration(self):
        if self._duration is None:
            return max(e.off for e in self)
        return float(self._duration)

    @property
    def sampling_rate(self):
        return self._sampling_rate

    @classmethod
    def from_bools(cls, bools, sampling_rate: float = DEFAULT_SAMPLING_RATE_HZ):
        """Converts the given boolean array into a series of discrete Events"""
        transitions = np.where(np.diff(bools))[0] + 1
        transitions = transitions.tolist()
        transitions = ([0] if bools[0] else []) + transitions + ([len(bools)] if bools[-1] else [])
        events = [Event(on, off) for on, off in zip(transitions[0::2], transitions[1::2])]
        return cls(events, duration=len(bools), sampling_rate=sampling_rate)

    def set_tolerances(self, vals):
        """Iteratively set the tolerances for all Events to the given vals"""
        for e in self:
            e.tolerances = vals

    def as_bools(self, round_method_on=np.ceil, round_method_off=np.floor):
        """Converts the Events to a boolean array, using parameterizable rounding methods"""
        bools = np.zeros(int(np.ceil(self.duration)), dtype=bool)
        for e in self:
            on, off = round_method_on(e.on).astype(int), round_method_off(e.off).astype(int)
            bools[on:off] = True
        return bools

    def as_dataframe(self):
        """Export events to a DataFrame"""
        return pd.DataFrame([(e.on, e.off) for e in self], columns=["on", "off"])

    def compute_agreement(self, other, normalize=False):
        """Computes an agreement matrix with the given EventSeries

        Parameters
        ----------
        other : EventSeries
        normalize : Optional(bool), default=False
            Normalize results rows by the number of events in the respective "reference" series

        Returns
        -------
        mat : pd.DataFrame
            2 x 2 agreement matrix
        """
        self._assert_is_compatible(other)
        mat = []
        series = [self, other]
        for ev in series:
            mat.append([sum(e in ev2 for e in ev) for ev2 in series])
        mat = pd.DataFrame(mat)
        mat.index = ["self (ref)", "other (ref)"]
        mat.columns = ["self", "other"]
        if normalize:
            mat = (mat.T / [len(ev) for ev in series]).T
        return mat

    @staticmethod
    def _format_confusion_matrix(mat):
        """Convenience function to convert to a DataFrame"""
        mat = pd.DataFrame(mat)
        mat.index = ["self (N)", "self (P)"]
        mat.columns = ["other (N)", "other (P)"]
        return mat

    def event_confusion_matrix(self, other, normalize=None, assimilate_events=True):
        """Generates a confusion matrix with the given EventSeries via Event-wise logic

        Parameters
        ----------
        other : EventSeries
        normalize : Optional(bool), default=False
        assimilate_events : Optional(bool), default=True
            If true, any overlapping Events from either series will be counted as one.

        Returns
        -------
        mat : pd.DataFrame
        """
        self._assert_is_compatible(other)
        if assimilate_events:
            truth = [e in self for e in self + other]
            pred = [e in other for e in self + other]
        else:
            truth = [e in self for e in self] + [e in self for e in other]
            pred = [e in other for e in self] + [e in other for e in other]
        mat = confusion_matrix(truth, pred, normalize)
        return self._format_confusion_matrix(mat)

    def epoch_confusion_matrix(self, other, normalize=None):
        """Generates a confusion matrix with the given EventSeries epoch-by-epoch via boolean logic

        Parameters
        ----------
        other : EventSeries
        normalize : Optional(bool), default=False

        Returns
        -------
        mat : pd.DataFrame
        """
        self._assert_is_compatible(other)
        truth, pred = self.as_bools(), other.as_bools()
        mat = confusion_matrix(truth, pred, normalize)
        return self._format_confusion_matrix(mat)

    def plot(self, bottom=0, top=1, color=None, as_seconds=True):
        """Simple plot with events filled between the given y values"""
        color = Colors.BLUE1 if color is None else color
        hz = self.sampling_rate if as_seconds else 1
        for event in self:
            plt.fill_betweenx(
                [bottom, top], event.on / hz, event.off / hz, alpha=0.3, color=color, linewidth=0
            )
        plt.xlim([0, self.duration / hz])
        plt.xlabel(f"Time ({'seconds' if as_seconds else 'samples'})")


def confusion_matrix(true, pred, normalize=None):
    """
    Notes
    -----
    This is a paraphrased implementation of scikit-learn's. We can use that method,
    but this allows our footprint and dependencies to remain minimal.
    """
    assert len(true) == len(pred)
    true, pred = np.array(true), np.array(pred)
    classes = np.unique([true, pred])
    n = len(classes)
    mat = [sum((true == a) & (pred == b)) for a in classes for b in classes]
    mat = np.reshape(mat, (n, n))

    # normalize
    if normalize is None:
        return mat
    mapping = {"all": None, "pred": 0, "true": 1}
    if normalize not in mapping:
        raise ValueError("Value of 'normalize' not understood. Try one of {true, pred, all}")
    axis = mapping[normalize]
    counts = mat.sum(axis=axis).clip(1, None)
    counts = np.ones_like(mat) * np.atleast_2d(counts)
    if axis == 1:
        counts = counts.T
    return mat / counts


class EventStack(dict):
    KEY_META = "meta"
    KEY_EVENT_STACK = "event_stack"
    KEY_DURATION = "duration"
    KEY_SAMPLING_RATE = "sampling_rate"

    KEY_ASSIGNED_ENUM = "AssignedEnum"

    def __init__(self, event_stack: Dict[IntEnum, EventSeries]):
        """Ingests multiple concurrent EventSeries objects, each with a key

        Parameters
        ----------
        event_stack : Dict[IntEnum, EventSeries]
        """
        vals = event_stack.values()
        for item in vals:
            assert isinstance(item, EventSeries)
        self._duration = self._infer_duration(vals)
        self._sampling_rate = self._infer_sampling_rate(vals)
        for k, v in event_stack.items():
            self._check_compatible(k, v)
        super().__init__(event_stack)

    @classmethod
    def from_string_keys(cls, event_stack: Dict[str, EventSeries]):
        assigned_enum = IntEnum(cls.KEY_ASSIGNED_ENUM, list(event_stack), start=0)
        return cls(event_stack={assigned_enum[k]: v for k, v in event_stack.items()})

    @staticmethod
    def _infer_duration(event_stack: ValuesView[EventSeries]) -> float:
        """Finds the common duration of each EventSeries and raises exception if there are multiple durations"""
        durations = list(set(item.duration for item in event_stack))
        if len(durations) != 1:
            raise ValueError("Events of differing durations given.")
        return durations[0]

    @staticmethod
    def _infer_sampling_rate(event_stack: ValuesView[EventSeries]) -> float:
        """Finds the common sampling rate of each EventSeries and raises exception if there are multiple values"""
        sampling_rates = list(set(item.sampling_rate for item in event_stack))
        if len(sampling_rates) != 1:
            raise ValueError("Events of differing sampling rates given.")
        return sampling_rates[0]

    @property
    def duration(self):
        return self._duration

    @property
    def sampling_rate(self):
        return self._sampling_rate

    @property
    def classes(self):
        return sorted(set(self))

    def _check_compatible(self, key, value):
        """Makes sure the given key/value pair would be a valid entry for this EventStack"""
        assert value.duration == self.duration
        assert value.sampling_rate == self.sampling_rate
        assert isinstance(key, IntEnum)

    def to_dict(self) -> dict:
        meta = {self.KEY_DURATION: self.duration, self.KEY_SAMPLING_RATE: self.sampling_rate}
        stack = {k.name: [dict(a) for a in v] for k, v in self.items()}
        return {self.KEY_META: meta, self.KEY_EVENT_STACK: stack}

    @classmethod
    def from_dict(cls, data: dict, intenum: IntEnum) -> EventStack:
        meta, event_stack = data[cls.KEY_META], data[cls.KEY_EVENT_STACK]
        return cls(
            {
                intenum[k]: EventSeries([Event(**ev) for ev in v], **meta)
                for k, v in event_stack.items()
            }
        )

    def write_json(self, filepath):
        """Serializes and exports event data to a JSON file

        Notes
        -----
        We're not exporting any OverlapTolerances here...
        seems like the kind of thing you'd only want to do explicitly during analysis anyway
        """
        with open(filepath, "w") as f:
            json.dump(self.to_dict(), f, indent=2)

    @classmethod
    def read_json(cls, filepath, intenum: IntEnum):
        """Construct from the given filepath"""
        with open(filepath, "r") as f:
            return cls.from_dict(data=json.load(f), intenum=intenum)

    def set_tolerances(self, vals):
        """Iteratively set the tolerances of all Events within all EventSeries"""
        for e in self.values():
            e.set_tolerances(vals)

    def resample(self, sampling_rate):
        """Make a new EventStack with resampled EventSeries values"""
        cls = type(self)
        return cls({k: series.resample(sampling_rate) for k, series in self.items()})

    def trim(self, start: Optional[int] = None, end: Optional[int] = None):
        """Make a new, trimmed version of this EventStack"""
        cls = type(self)
        return cls({k: series.trim(start, end) for k, series in self.items()})

    def plot(self):
        """Stacks individual series onto the same plot"""
        n = max(self) + 1
        cmap = LinearSegmentedColormap.from_list("", ["black", Colors.BLUE1], N=n)
        fig, ax = plt.subplots(figsize=(10, 0.5 * len(self)))
        for k, v in self.items():
            v.plot(bottom=0, top=k, color=cmap(k))
        ax.set_yticks([k.value for k in self])
        ax.set_yticklabels([k.name for k in self])
        ax.set_ylim(ax.get_ylim())
        return fig


class EventClassSeries(EventStack):
    """A special type of EventStack wherein every timestamp is expected to have one and only one class

    Notes
    -----
    Useful for performing class confusion or agreement analytic, such as with hypnogram annotations
    """

    def __init__(
        self, event_stack: Dict[IntEnum, EventSeries], missing_enum: Optional[IntEnum] = None
    ):
        """Ingests multiple concurrent EventSeries objects, each with a key

        Parameters
        ----------
        event_stack : Dict[IntEnum, EventSeries]
        missing_enum : IntEnum, Optional
            value to fill where no annotation exists
        """
        vals = event_stack.values()
        for item in vals:
            assert isinstance(item, EventSeries)
        if missing_enum is not None:
            missing = np.sum([s.as_bools() for s in vals], axis=0) == 0
            event_stack[missing_enum] = EventSeries.from_bools(
                missing, sampling_rate=self._infer_sampling_rate(vals)
            )
        msg = "Input annotations must be continuous and flat!"
        assert self._is_continuous_and_flat(vals), msg
        super().__init__(event_stack)

    @classmethod
    def _is_continuous_and_flat(cls, event_stack: ValuesView[EventSeries]):
        """One and only one class active at each timestamp

        Returns
        -------
        is_continuous_and_flat : bool
        """
        events = sorted([ev for v in event_stack for ev in v], key=lambda x: x.on)
        starts_at_zero = events[0].on == 0
        ends_at_duration = events[-1].off == cls._infer_duration(event_stack)
        is_continuous_and_flat = all(
            [events[i].abuts(events[i + 1]) for i in range(len(events) - 1)]
        )
        return is_continuous_and_flat and starts_at_zero and ends_at_duration

    @classmethod
    def from_array(cls, class_array, sampling_rate: float = DEFAULT_SAMPLING_RATE_HZ):
        """Initialize from a 1-D array of classes"""
        assert isinstance(class_array, (list, np.ndarray, pd.Series))
        assert all(isinstance(c, IntEnum) for c in class_array)
        events = dict()
        for c in sorted(set(class_array)):
            bools = [a == c for a in class_array]
            events[c] = EventSeries.from_bools(bools, sampling_rate=sampling_rate)
        return cls(events)

    def as_array(self, round_method_on=np.ceil, round_method_off=np.floor):
        """Export all events as a list of classes"""
        array = [None] * int(round_method_off(self.duration))
        for k, events in self.items():
            bools = events.as_bools(round_method_on, round_method_off)
            array = [k if b else a for a, b in zip(array, bools)]
        if None in array:
            raise RuntimeError(
                "None detected in result of EventClassSeries.as_array, this is a bug"
            )
        return array

    def event_confusion_matrix(self, other, assimilate_events=True):
        """Generate a class-wise confusion matrix with 'other'

        Parameters
        ----------
        other : EventClassSeries
        assimilate_events : Optional(bool), default=True
            If true, any overlapping Events from either series will be counted as one.

        Returns
        -------
        mat : pd.DataFrame
        """
        mat = []
        for s in self.classes:
            events_self = self[s]
            for o in self.classes:
                events_other = other[o]
                if assimilate_events:
                    truth = [e in events_self for e in events_self + events_other]
                    pred = [e in events_other for e in events_self + events_other]
                else:
                    truth = [e in events_self for e in events_self] + [
                        e in events_self for e in events_other
                    ]
                    pred = [e in events_other for e in events_self] + [
                        e in events_other for e in events_other
                    ]
                mat.append(sum(np.array(truth) & np.array(pred)))
        mat = np.reshape(mat, (len(self), len(self)))
        return self._format_confusion_matrix(mat)

    def epoch_confusion_matrix(self, other, normalize=None):
        """Generates a confusion matrix with 'other', epoch-by-epoch via boolean logic

        Parameters
        ----------
        other : EventClassSeries
        normalize : Optional(bool), default=False

        Returns
        -------
        mat : pd.DataFrame
        """
        truth, pred = self.as_array(), other.as_array()
        mat = confusion_matrix(truth, pred, normalize)
        return self._format_confusion_matrix(mat, labels=set(truth) | set(pred))

    def _format_confusion_matrix(self, mat, labels=None):
        """Convenience function to convert confusion matrix to a DataFrame"""
        mat = pd.DataFrame(mat)
        labels = self.classes if labels is None else labels
        mat.index = [f"self ({c.name})" for c in labels]
        mat.columns = [f"other ({c.name})" for c in labels]
        return mat


class SeizureEvent(Event):
    default_duration = 120


class MissingDataLabels(IntEnum):
    MISSING = 0
    UNSCORABLE = -1
    ARTIFACT = -2
    DISAGREE = -3



#!/usr/bin/env python

import tempfile
from pathlib import Path

import pytest

from bioevents import event_handling, hypnogram

from . import utils as test_utils


@pytest.fixture(scope="function")
def dummy_hypnogram():
    return test_utils.simulate_hypnogram(series_duration=100, num_cycles=5, avg_duration=5)


def test_sleep_metrics():
    stages = hypnogram.SleepStages

    # create a dummy hypnogram
    hypno = hypnogram.Hypnogram.from_array(
        [stages.W] * 90
        + [stages.N1] * 30
        + [stages.W] * 30
        + [stages.N1] * 30
        + [stages.N2] * 60
        + [stages.N3] * 90
        + [stages.REM] * 30
        + [stages.W] * 60
    )
    # ensure the total sleep time is the sum of total non-wake stages in minutes
    assert hypno.total_recording_time() == 420
    assert hypno.total_sleep_time() == 240
    assert hypno.sleep_efficiency_percent() == (240 / 420) * 100
    assert hypno.wake_after_sleep_onset() == 90

    # test functions given bedtime_series
    bedtime_series = event_handling.EventSeries(
        events=[event_handling.Event(30, 390)],
        duration=hypno.duration,
    )
    assert hypno.total_recording_time(bedtime_series=bedtime_series) == 360
    assert hypno.total_sleep_time(bedtime_series=bedtime_series) == 240
    assert hypno.sleep_efficiency_percent(bedtime_series=bedtime_series) == 240 / 360 * 100


def test_time_in_stage():
    stages = [hypnogram.SleepStages.W] * 60 * 20
    stages.extend([hypnogram.SleepStages.N1] * 60 * 10)
    stages.extend([hypnogram.SleepStages.REM] * 60 * 10)
    stages.extend([hypnogram.SleepStages.W] * 60 * 60)
    hypno = hypnogram.Hypnogram.from_array(stages)
    assert hypno.time_in_stage(hypnogram.SleepStages.N1) == 60 * 10
    assert hypno.time_in_stage(hypnogram.SleepStages.REM) == 60 * 10
    assert hypno.time_in_stage(hypnogram.SleepStages.N2) == 0
    assert hypno.time_in_stage(hypnogram.SleepStages.N3) == 0


@pytest.mark.parametrize("n1_tolerance", [0, 60])
def test_sleep_onset_latency(n1_tolerance):
    # TODO refactor this test to multiple atomic ones
    stages = hypnogram.SleepStages

    # define two N1 events that we expect to treat differently when excluding sub-threshold N1 activity
    n1_exclude, n1_include = [stages.N1] * 60, [stages.N1] * 90

    # create a dummy hypnogram
    hypno = hypnogram.Hypnogram.from_array(
        [stages.W] * 120 + n1_exclude + [stages.W] * 30 + n1_include + [stages.W] * 90
    )

    # make sure the first N1 detection is flagged when including all N1 activity by default
    # make sure the first N1 detection is ignored when excluding short N1 activity
    ref = 210 if n1_tolerance else 120
    dut = hypno.sleep_onset_latency(n1_tolerance)
    assert dut == ref

    # create a dummy hypnogram with a non-N1 sleep stage
    hypno = hypnogram.Hypnogram.from_array(
        [stages.W] * 120
        + n1_exclude
        + [stages.W] * 30
        + n1_exclude
        + [stages.N2] * 30
        + [stages.W] * 90
    )

    # make sure the first N1 detection is flagged when including all N1 activity by default
    # with tolerance, make sure the second, short N1 is actually included if followed by a non-N1 sleep stage
    ref = 210 if n1_tolerance else 120
    dut = hypno.sleep_onset_latency(n1_tolerance)
    assert dut == ref

    # make sure that, if resampled, the result is consistent with the new units
    sampling_rate = 1 / 30
    resampled = hypno.resample(sampling_rate=sampling_rate)
    dut = resampled.sleep_onset_latency(n1_tolerance=n1_tolerance * sampling_rate)
    assert dut == ref * sampling_rate

    # create a dummy hypnogram with no sleep
    hypno = hypnogram.Hypnogram.from_array([stages.W] * 120)

    # make sure we get None if there is no sleep whatsoever
    ref = None
    dut = hypno.sleep_onset_latency(n1_tolerance)
    assert dut == ref

    hypno = hypnogram.Hypnogram.from_array([stages.W] * 120 + n1_exclude + [stages.W] * 30)
    # make sure we get None if there is sub-threshold N1
    ref = None if n1_tolerance else 120
    dut = hypno.sleep_onset_latency(n1_tolerance)
    assert dut == ref


def test_hypnogram_waso():
    # a hypnogram with sleep should have float-valued WASO
    wake_epochs = [hypnogram.SleepStages.W] * 10
    sleep_epochs = [hypnogram.SleepStages.REM] * 10
    dummy_hypnogram = hypnogram.Hypnogram.from_array(wake_epochs + sleep_epochs)
    WASO = dummy_hypnogram.wake_after_sleep_onset()
    assert isinstance(WASO, float)

    # a hypnogram with no sleep should have no WASO (i.e., None)
    dummy_hypnogram = hypnogram.Hypnogram.from_array(wake_epochs)
    WASO = dummy_hypnogram.wake_after_sleep_onset()
    assert WASO is None


@pytest.mark.parametrize("with_disagreement", [False, True])
def test_hypnogram_read_write_json(dummy_hypnogram, with_disagreement):
    if with_disagreement:
        # knock out a few samples
        arr = dummy_hypnogram.as_array()
        arr[5:9] = [hypnogram.Labels.DISAGREE] * 3
        dummy_hypnogram = hypnogram.Hypnogram.from_array(arr)
    with tempfile.TemporaryDirectory() as td:
        tf = Path(td) / "hypnogram.json"
        dummy_hypnogram.write_json(tf)
        dut = hypnogram.Hypnogram.read_json(tf)
    assert dut.duration == dummy_hypnogram.duration
    assert dut.classes == dummy_hypnogram.classes
    # note: this will iteratively compare the Events, which are in a list which is in this dict
    assert dut == dummy_hypnogram


def test_latency_rem_no_sleep():
    num_seconds = 60 * 100
    stages = [hypnogram.SleepStages.W] * num_seconds
    hypno = hypnogram.Hypnogram.from_array(stages)
    assert hypno.latency_to_stage(hypnogram.SleepStages.REM) is None


def test_latency_rem_no_rem():
    # simulate 100 minutes of sleep with persistent sleep but no REM
    stages = [hypnogram.SleepStages.W] * 60 * 20
    stages.extend([hypnogram.SleepStages.N1] * 60 * 10)
    stages.extend([hypnogram.SleepStages.N2] * 60 * 10)
    stages.extend([hypnogram.SleepStages.W] * 60 * 60)
    hypno = hypnogram.Hypnogram.from_array(stages)

    assert hypno.latency_to_stage(hypnogram.SleepStages.REM) is None


def test_latency_to_stage():
    stages = [hypnogram.SleepStages.W] * 20
    stages.extend([hypnogram.SleepStages.N1] * 10)
    stages.extend([hypnogram.SleepStages.N2] * 10)
    stages.extend([hypnogram.SleepStages.REM] * 10)
    hypno = hypnogram.Hypnogram.from_array(stages)
    assert hypno.latency_to_stage(hypnogram.SleepStages.N1) == 20
    assert hypno.latency_to_stage(hypnogram.SleepStages.N2) == 30
    assert hypno.latency_to_stage(hypnogram.SleepStages.REM) == 40
    assert hypno.latency_to_stage(hypnogram.SleepStages.N3) is None


def test_hypnogram_plot_no_errors(dummy_hypnogram):
    """A rudimentary check that we hit no errors when building a Hypnogram plot"""
    dummy_hypnogram.plot()


def test_hypnogram_report_no_errors(dummy_hypnogram):
    report = dummy_hypnogram.generate_report()



import tempfile
from enum import IntEnum
from pathlib import Path

import numpy as np
import pytest

from bioevents import event_handling, hypnogram

from .utils import simulate_hypnogram


def test_overlap_tolerance_setattr():
    illegal_param = -1
    # test setattr via init
    with pytest.raises(AssertionError):
        event_handling.OverlapTolerances(diff_on=illegal_param)
    tol = event_handling.OverlapTolerances()
    with pytest.raises(AssertionError):
        tol.diff_on = illegal_param


def test_event_init_on_only():
    on = 20
    event = event_handling.Event(on=on)
    assert event.on == on
    assert event.duration == event.default_duration
    assert event.off == on + event.default_duration


def test_event_init():
    on, off = 20, 35
    event = event_handling.Event(on=on, off=off)
    assert event.on == on
    assert event.off == off
    assert event.duration == off - on


def test_event_init_invalid():
    off = 10
    on_after_off = 12
    on_negative = -1
    for on in [on_after_off, on_negative]:
        with pytest.raises(AssertionError):
            event_handling.Event(on=on, off=off)


def test_event_from_duration():
    on, duration = 20, 15
    event = event_handling.Event.from_duration(on=on, duration=duration)
    assert event.on == on
    assert event.off == on + duration
    assert event.duration == duration


def test_event_abuts():
    event_a = event_handling.Event(on=10, off=20)
    assert event_handling.Event(on=20, off=30).abuts(event_a)
    assert event_handling.Event(on=5, off=10).abuts(event_a)
    assert not event_handling.Event(on=5, off=10.5).abuts(event_a)
    assert not event_handling.Event(on=19, off=30).abuts(event_a)
    assert not event_handling.Event(on=12, off=18).abuts(event_a)
    assert not event_handling.Event(on=8, off=22).abuts(event_a)
    assert not event_handling.Event(on=3, off=8).abuts(event_a)
    assert not event_handling.Event(on=22, off=29).abuts(event_a)


def test_event_overlaps_default():
    event_a = event_handling.Event(on=10, off=20)

    # test clearly overlapping events overlap under default conditions
    event_b = event_handling.Event(on=15, off=30)
    assert event_a.overlaps(event_b)
    assert event_b.overlaps(event_a)

    # test directly adjacent events overlap under default conditions
    event_b = event_handling.Event(on=20, off=30)
    assert event_a.overlaps(event_b)
    assert event_b.overlaps(event_a)

    # test totally separate events do not overlap
    event_b = event_handling.Event(on=21, off=30)
    assert not event_a.overlaps(event_b)
    assert not event_b.overlaps(event_a)

    # test event fitting inside other event will overlap
    event_b = event_handling.Event(on=12, off=18)
    assert event_a.overlaps(event_b)
    assert event_b.overlaps(event_a)

    # test identical events overlap
    event_b = event_a.__copy__()
    assert event_a.overlaps(event_b)
    assert event_b.overlaps(event_a)


def test_event_overlaps_diffs():
    event_a = event_handling.Event(on=10, off=20)
    event_b = event_handling.Event(on=12, off=22)

    # test events overlap with acceptable on diff
    tol = event_handling.OverlapTolerances(diff_on=5)
    event_a.tolerances = tol
    assert event_a.overlaps(event_b)

    # test events overlap with acceptable off diff
    tol = event_handling.OverlapTolerances(diff_off=5)
    event_a.tolerances = tol
    assert event_a.overlaps(event_b)

    # test events overlap with maximum acceptable on diff
    tol = event_handling.OverlapTolerances(diff_on=2)
    event_a.tolerances = tol
    assert event_a.overlaps(event_b)

    # test events overlap with maximum acceptable off diff
    tol = event_handling.OverlapTolerances(diff_off=2)
    event_a.tolerances = tol
    assert event_a.overlaps(event_b)

    # test events do not overlap with unacceptable on diff
    tol = event_handling.OverlapTolerances(diff_on=1)
    event_a.tolerances = tol
    assert not event_a.overlaps(event_b)
    assert event_b.overlaps(event_a)  # event_b should be unaffected by updated tolerances

    # test events do not overlap with unacceptable off diff
    tol = event_handling.OverlapTolerances(diff_off=1)
    event_a.tolerances = tol
    assert not event_a.overlaps(event_b)
    assert event_b.overlaps(event_a)  # event_b should be unaffected by updated tolerances


def test_event_overlaps_ratios():
    event_a = event_handling.Event(on=10, off=20)
    event_b = event_handling.Event(on=12, off=22)

    # test events overlap with acceptable on ratio
    tol = event_handling.OverlapTolerances(ratio_on=0.5)
    event_a.tolerances = tol
    assert event_a.overlaps(event_b)

    # test events overlap with acceptable off ratio
    tol = event_handling.OverlapTolerances(ratio_off=0.5)
    event_a.tolerances = tol
    assert event_a.overlaps(event_b)

    # test events overlap with maximum acceptable on ratio
    tol = event_handling.OverlapTolerances(ratio_on=0.2)
    event_a.tolerances = tol
    assert event_a.overlaps(event_b)

    # test events overlap with maximum acceptable off ratio
    tol = event_handling.OverlapTolerances(ratio_off=0.2)
    event_a.tolerances = tol
    assert event_a.overlaps(event_b)

    # test events do not overlap with unacceptable on ratio
    tol = event_handling.OverlapTolerances(ratio_on=0.1)
    event_a.tolerances = tol
    assert not event_a.overlaps(event_b)
    assert event_b.overlaps(event_a)  # event_b should be unaffected by updated tolerances

    # test events do not overlap with unacceptable off ratio
    tol = event_handling.OverlapTolerances(ratio_off=0.1)
    event_a.tolerances = tol
    assert not event_a.overlaps(event_b)
    assert event_b.overlaps(event_a)  # event_b should be unaffected by updated tolerances


@pytest.mark.parametrize(
    "on, bools",
    [
        (2, [False, False, True, True, False]),  # event is contained within series
        (2, [False, False, True, True]),  # event is clipped by series end
        (0, [True, True, False]),  # event is clipped by series start
    ],
)
def test_event_series_from_bools_clipped(bools, on):
    dut = event_handling.EventSeries.from_bools(bools)

    # we're expecting one event
    assert len(dut) == 1

    # check the interpreted timestamps
    event = dut[0]
    assert event.on == on
    assert event.off == on + 2


@pytest.mark.parametrize(
    "on, off, duration, ref",
    [
        (2, 4, 5, [False, False, True, True, False]),  # event is contained within series
        (2, 4, 4, [False, False, True, True]),  # event is clipped by series end
        (0, 2, 3, [True, True, False]),  # event is clipped by series start
    ],
)
def test_event_series_as_bools_clipped(on, off, duration, ref):
    series = event_handling.EventSeries(events=[event_handling.Event(on, off)], duration=duration)
    dut = series.as_bools()
    np.testing.assert_equal(ref, dut)


def test_event_series_from_bools():
    """Define two events in boolean form and make sure they're ingested correctly as events"""
    bools = [False, False, True, True, False, True, False, False]
    series = event_handling.EventSeries.from_bools(bools)
    assert len(series) == 2
    assert series[0].on == 2
    assert series[0].off == 4
    assert series[1].on == 5
    assert series[1].off == 6


def test_event_series_as_bools():
    """Define two events in Event form and make sure they're exported correctly in boolean form"""
    ref = [False, False, True, True, False, True, False, False]
    events = [
        event_handling.Event(on=2, off=4),
        event_handling.Event(on=5, off=6),
    ]
    series = event_handling.EventSeries(events, duration=len(ref))
    dut = series.as_bools()
    np.testing.assert_equal(ref, dut)


@pytest.mark.parametrize("sampling_rate", [0.5, 1.0, 10.0])
def test_event_series_resample(sampling_rate):
    original_sampling_rate = 1.0
    # make a dummy event array and resample it
    ref = event_handling.EventSeries(
        [
            event_handling.Event(0, 1),
            event_handling.Event(3, 6),
            event_handling.Event(7.5, 8.5),
            event_handling.Event(9, 11),
            event_handling.Event(10, 12),
            event_handling.Event(14, 20),
        ],
        duration=30,
        sampling_rate=original_sampling_rate,
    )
    dut = ref.resample(sampling_rate=sampling_rate)

    # make sure the sample duration is adjusted correctly
    np.testing.assert_allclose(dut.duration / sampling_rate, ref.duration)

    # make sure we end up with the same number of events
    assert len(dut) == len(ref)

    # make sure each timestamp is accurately resampled
    for d, r in zip(dut, ref):
        assert d.on / sampling_rate == r.on

    # resample back to original rate
    dut = dut.resample(sampling_rate=original_sampling_rate)

    # make sure the sample duration is adjusted correctly
    np.testing.assert_allclose(dut.duration / original_sampling_rate, ref.duration)


@pytest.mark.parametrize("start", [0, 1, 1.5])
def test_event_series_trim(start):
    # make a dummy event array
    series = event_handling.EventSeries(
        [
            event_handling.Event(
                0, 1
            ),  # this event should be eliminated entirely unless start == 0
            event_handling.Event(3, 6),  # this event should remain the same
            event_handling.Event(7.5, 8.5),  # just testing float values
            event_handling.Event(9, 11),  # this event should be cut in half
            event_handling.Event(10, 12),  # this event should be eliminated entirely
            event_handling.Event(14, 20),  # this event should be eliminated entirely
        ],
        duration=30,
    )

    # make sure we raise an error with parameters out of bounds
    with pytest.raises(ValueError):
        series.trim(-1, 10)
    with pytest.raises(ValueError):
        series.trim(10, 40)

    # trim our boolean array
    expected = event_handling.EventSeries(
        [
            event_handling.Event(3 - start, 6 - start),
            event_handling.Event(7.5 - start, 8.5 - start),
            event_handling.Event(9 - start, 10 - start),
        ],
        duration=10 - start,
    )
    if start < 1:
        expected.append(event_handling.Event(0, 1))
    expected._resolve_events()
    actual = series.trim(start, 10)

    assert actual.duration == expected.duration
    assert len(actual) == len(expected)

    for act, exp in zip(actual, expected):
        assert act == exp


@pytest.mark.parametrize(
    "duration, events_original, events_expected",
    [
        (10, [], [event_handling.Event(0, 10)]),
        (
            10,
            [event_handling.Event(0.5, 2), event_handling.Event(4, 5)],
            [
                event_handling.Event(0, 0.5),
                event_handling.Event(2, 4),
                event_handling.Event(5, 10),
            ],
        ),
        (
            10,
            [event_handling.Event(0, 2), event_handling.Event(4, 5)],
            [event_handling.Event(2, 4), event_handling.Event(5, 10)],
        ),
        (
            5,
            [event_handling.Event(0.5, 2), event_handling.Event(4, 5)],
            [event_handling.Event(0, 0.5), event_handling.Event(2, 4)],
        ),
        (
            10,
            [event_handling.Event(0.5, 2)],
            [event_handling.Event(0, 0.5), event_handling.Event(2, 10)],
        ),
    ],
)
def test_event_series_invert(duration, events_original, events_expected):
    original = event_handling.EventSeries(events=events_original, duration=duration)
    expected = event_handling.EventSeries(events=events_expected, duration=duration)
    inverted = ~original
    assert inverted.duration == original.duration
    assert inverted == expected


def test_event_class_series_missing_enum():
    class MyIntEnum(IntEnum):
        A = 0
        B = 1
        C = 2

    event_stack = {
        MyIntEnum.A: event_handling.EventSeries.from_bools([False, True, False]),
        MyIntEnum.B: event_handling.EventSeries.from_bools([True, False, False]),
    }

    # make sure the missing bin is rejected if no missing enum arg is given
    with pytest.raises(AssertionError):
        event_handling.EventClassSeries(event_stack)

    # make sure the missing bin is filled with the missing enum arg given
    dut = event_handling.EventClassSeries(event_stack, missing_enum=MyIntEnum.C)
    assert dut.as_array() == [MyIntEnum.B, MyIntEnum.A, MyIntEnum.C]


def test_write_read_json():
    ref = simulate_hypnogram(series_duration=100)
    ref = ref.resample(2)
    with tempfile.TemporaryDirectory() as td:
        fp = Path(td) / "temp.json"
        ref.write_json(fp)
        dut = event_handling.EventClassSeries.read_json(fp, hypnogram.SleepStages)

    assert ref.classes == dut.classes
    assert ref.duration == dut.duration
    assert ref.sampling_rate == dut.sampling_rate
    assert ref.as_array() == dut.as_array()


@pytest.mark.parametrize(
    "on, duration, expected",
    [
        (4, 10, True),  # this is the only valid one
        (4, 20, False),  # no events for 10-20
        (3, 10, False),
        (3.9, 10, False),
        (4.1, 10, False),
        (5, 10, False),
    ],
)
def test_event_stack_is_continuous_and_flat(on, duration, expected):
    func = event_handling.EventClassSeries._is_continuous_and_flat

    # note that events are not in order -- this shouldn't be a problem...
    events_a = event_handling.EventSeries(
        events=[event_handling.Event(3, 4), event_handling.Event(1, 3)], duration=duration
    )
    events_b = event_handling.EventSeries(
        events=[event_handling.Event(on, 6), event_handling.Event(6, 10)], duration=duration
    )
    stack = {"a": events_a, "b": events_b}

    # this stack doesn't start at 0 -- this SHOULD be a problem
    assert not func(stack.values())

    # add an event at 0 to make it continuous
    stack["c"] = event_handling.EventSeries([event_handling.Event(0, 1)], duration=duration)
    assert func(stack.values()) == expected


def test_event_stack_from_string_keys():
    event_stack = {
        "b": event_handling.EventSeries.from_bools([False, True, True, False]),
        "a": event_handling.EventSeries.from_bools([False, True, True, False]),
    }
    dut = event_handling.EventStack.from_string_keys(event_stack)

    # make sure an IntEnum subclass was assigned
    assert all(isinstance(k, IntEnum) for k in dut)

    # make sure elements were assigned to a zero-indexed enum
    assigned_enum = type(list(dut.keys())[0])
    assert assigned_enum["b"] == 0
    assert assigned_enum.__name__ == event_handling.EventStack.KEY_ASSIGNED_ENUM


def test_event_series_append():
    # make a dummy boolean event array
    events = [
        event_handling.Event(0, 1),
        event_handling.Event(3, 6),
        event_handling.Event(7.5, 8.5),
        event_handling.Event(9, 11),
        event_handling.Event(10, 12),
        event_handling.Event(14, 20),
    ]
    duration = 30

    # test appending each individual event in turn
    ref = event_handling.EventSeries(events, duration)
    for event in events:
        dut = event_handling.EventSeries(
            events=[ev for ev in events if ev != event], duration=duration
        )
        dut.append(event)
        assert len(ref) == len(dut), "Reference and test have different number of events!"

        # note: this will also check the order of the events!
        assert ref == dut, "Reference events do not match test events!"


def test_event_series_extend():
    # make a dummy boolean event array
    events = [
        event_handling.Event(0, 1),
        event_handling.Event(7.5, 8.5),
        event_handling.Event(9, 11),
        event_handling.Event(10, 12),
    ]
    duration = 30

    extend_events = [
        event_handling.Event(3, 6),
        event_handling.Event(14, 20),
    ]

    # test appending each individual event in turn
    ref = event_handling.EventSeries(events + extend_events, duration)
    dut = event_handling.EventSeries(events=events, duration=duration)
    dut.extend(extend_events)
    assert len(ref) == len(dut), "Reference and test have different number of events!"

    # note: this will also check the order of the events!
    assert ref == dut, "Reference events do not match test events!"


def test_event_series_debounce():
    # metadata we need to preserve
    duration = 30
    sampling_rate = 2

    # debouncing parameters
    persistence_period = 1.5
    min_duration = 3.5

    # make a dummy boolean event array
    events = [
        event_handling.Event(0, 3.5),
        event_handling.Event(5.5, 6),  # too short -- should be pruned away
        event_handling.Event(8, 9),  # won't duration test, but should be merged w following
        event_handling.Event(10.25, 15),
    ]

    # make the expected debounced events based on the parameters above
    events_debounced = [
        event_handling.Event(0, 3.5),
        event_handling.Event(8, 15),
    ]

    original = event_handling.EventSeries(events, duration=duration, sampling_rate=sampling_rate)
    expected = event_handling.EventSeries(
        events_debounced, duration=duration, sampling_rate=sampling_rate
    )
    actual = original.debounce(persistence_period=persistence_period, min_duration=min_duration)
    assert actual.sampling_rate == expected.sampling_rate
    assert actual.duration == expected.duration
    assert actual == expected
